/**
dco/c++ v3.7.1
    -- Algorithmic Differentiation by Operator Overloading in C++

COPYRIGHT 2022
The Numerical Algorithms Group Limited and
Software and Tools for Computational Engineering @ RWTH Aachen University

This file is part of dco/c++.
**/

#include <cstdlib>
#include <iostream>
#include <cassert>
#include "dco.hpp"
using namespace std;
using namespace dco;
typedef ga1s<double> DCO_MODE;
typedef DCO_MODE::type DCO_TYPE;
typedef DCO_MODE::tape_t DCO_TAPE_TYPE;

#include "../include/f.hpp"

template <typename TYPE> void EulerSteps(const size_t m, const vector<TYPE>& A, vector<TYPE>& T) {
  size_t n = T.size();
  for (size_t j = 0; j < m; j++) {
    for (size_t i = 0; i < n; ++i)
      T[i] = -T[i];
    Solve(A, T);
  }
}

template <typename DCO_TYPE> void EulerSteps_make_gap(const size_t m, const vector<DCO_TYPE>& A, vector<DCO_TYPE>& T) {
  typedef mode<DCO_TYPE> MODE_T;
  size_t n = T.size();

  auto T_in(T);

  // forward run
  MODE_T::global_tape->switch_to_passive();
  EulerSteps(m, A, T);
  MODE_T::global_tape->switch_to_active();

  // register output
  for (size_t i = 0; i < n; i++)
    MODE_T::global_tape->register_variable(T[i]);

  auto fill_gap = [=, &A]() {
    auto p = MODE_T::global_tape->get_position();
    vector<DCO_TYPE> Tl(T_in);

    // forward run
    EulerSteps(m, A, Tl);
    cerr << "record =" << dco::size_of(MODE_T::global_tape) << "B" << endl;

    // get output adjoints (seeds for this section)
    for (size_t i = 0; i < n; i++) {
      MODE_T::global_tape->register_output_variable(Tl[i]);
      dco::derivative(Tl[i]) = dco::derivative(T[i]);
    }

    // reverse run and reset to position p (so only reverse run of Solve)
    MODE_T::global_tape->interpret_adjoint_and_reset_to(p);
  };

  MODE_T::global_tape->insert_callback(std::move(fill_gap));
}

template <typename TYPE> inline void sim(const vector<TYPE>& c, const size_t m, vector<TYPE>& T, const size_t cs) {
  size_t n = T.size();
  static vector<TYPE> A(n * n); // static avoids copy
  residual_jacobian(c, T, A);
  for (size_t i = 0; i < n; ++i) {
    for (size_t j = 0; j < n; ++j)
      A[i * n + j] = A[i * n + j] / m;
    A[i + i * n] = A[i + i * n] - 1;
  }
  LUDecomp(A);
  for (size_t j = 0; j < m; j += cs) {
    size_t s = (j + cs < m) ? cs : m - j;
    EulerSteps_make_gap<DCO_TYPE>(s, A, T);
  }
}

template <typename TYPE, typename OTYPE>
inline void f(const vector<TYPE>& c, const size_t m, vector<TYPE>& T, const vector<OTYPE>& O, TYPE& v,
              const size_t cs) {
  sim(c, m, T, cs);
  v = 0;
  size_t n = T.size();
  for (size_t i = 0; i < n - 1; ++i)
    v = v + (T[i] - O[i]) * (T[i] - O[i]);
  v = v / (n - 1);
}

int main(int argc, char* argv[]) {
  if (argc != 4) {
    cerr << "2 parameters expected:" << endl
         << "  1. number of spatial finite difference grid points" << endl
         << "  2. number of implicit Euler steps" << endl
         << "  3. distance between checkpoints" << endl;
    return -1;
  }
  size_t n = atoi(argv[1]), m = atoi(argv[2]), cs = atoi(argv[3]);
  assert(cs <= m);
  vector<double> c(n);
  for (size_t i = 0; i < n; i++)
    c[i] = 0.01;
  vector<double> T(n);
  for (size_t i = 0; i < n - 1; i++)
    T[i] = 300.;
  T[n - 1] = 1700.;
  ifstream ifs("O.txt");
  vector<double> O(n);
  for (size_t i = 0; i < n; i++)
    ifs >> O[i];
  vector<double> dvdc(n);

  vector<DCO_TYPE> ca(n);
  vector<DCO_TYPE> Ta(n);
  DCO_TYPE va;
  DCO_MODE::global_tape = DCO_TAPE_TYPE::create();
  for (size_t i = 0; i < n; i++) {
    ca[i] = c[i];
    Ta[i] = T[i];
    DCO_MODE::global_tape->register_variable(ca[i]);
    DCO_MODE::global_tape->register_variable(Ta[i]);
  }
  f(ca, m, Ta, O, va, cs);
  cerr << "record=" << dco::size_of(DCO_MODE::global_tape) << "B" << endl;
  derivative(va) = 1.0;
  DCO_MODE::global_tape->register_output_variable(va);
  DCO_MODE::global_tape->interpret_adjoint();
  for (size_t i = 0; i < n; i++)
    dvdc[i] = derivative(ca[i]);
  DCO_TAPE_TYPE::remove(DCO_MODE::global_tape);
  cout.precision(5);
  for (size_t i = 0; i < n; i++)
    cout << "dvdc[" << i << "]=" << dvdc[i] << endl;
  return 0;
}
